use std::{
    any::Any,
    marker::PhantomData,
    ptr::{self, NonNull},
    sync::{
        Arc,
        atomic::{AtomicPtr, Ordering},
    },
};

use async_task::Runnable;

use crate::{
    context::Context,
    job::{HeapJob, Job},
    latch::{AsCoreLatch, CountLatch, MutexLatch, WakeLatch},
    util::{DropGuard, SendPtr},
    workerthread::WorkerThread,
};

pub struct Scope<'scope, 'env: 'scope> {
    // latch to wait on before the scope finishes
    job_counter: CountLatch<MutexLatch>,
    // local threadpool
    context: Arc<Context>,
    // panic error
    panic: AtomicPtr<Box<dyn Any + Send + 'static>>,
    // variant lifetime
    _scope: PhantomData<&'scope mut &'scope ()>,
    _env: PhantomData<&'env mut &'env ()>,
}

pub fn scope<'env, F, R>(f: F) -> R
where
    F: for<'scope> FnOnce(&'scope Scope<'scope, 'env>) -> R + Send,
    R: Send,
{
    scope_with_context(Context::global_context(), f)
}

pub fn scope_with_context<'env, F, R>(context: &Arc<Context>, f: F) -> R
where
    F: for<'scope> FnOnce(&'scope Scope<'scope, 'env>) -> R + Send,
    R: Send,
{
    context.run_in_worker(|worker| {
        // SAFETY: we call complete() after creating this scope, which
        // ensures that any jobs spawned from the scope exit before the
        // scope closes.
        let this = unsafe { Scope::from_context(context.clone()) };
        this.complete(worker, || f(&this))
    })
}

impl<'scope, 'env> Scope<'scope, 'env> {
    fn wait_for_jobs(&self, worker: &WorkerThread) {
        if self.job_counter.count() > 0 {
            tracing::trace!("waiting for {} jobs to finish.", self.job_counter.count());
            tracing::trace!("thread id: {:?}, jobs: {:?}", worker.index, unsafe {
                worker.queue.as_ref_unchecked()
            });

            // set worker index in the job counter
            worker.wait_until_latch(self.job_counter.as_core_latch());
        }
    }

    /// should be called from within a worker thread.
    fn complete<F, R>(&self, worker: &WorkerThread, f: F) -> R
    where
        F: FnOnce() -> R + Send,
        R: Send,
    {
        use std::panic::{AssertUnwindSafe, catch_unwind};

        #[allow(dead_code)]
        fn make_job<F: FnOnce() -> T, T>(f: F) -> Job<T> {
            #[align(8)]
            unsafe fn harness<F: FnOnce() -> T, T>(this: *const (), job: *const Job<T>) {
                let f = unsafe { Box::from_raw(this.cast::<F>().cast_mut()) };

                let result = catch_unwind(AssertUnwindSafe(move || f()));

                let job = unsafe { Box::from_raw(job.cast_mut()) };
                job.complete(result);
            }

            Job::<T>::new(harness::<F, T>, unsafe {
                NonNull::new_unchecked(Box::into_raw(Box::new(f))).cast()
            })
        }

        let result = match catch_unwind(AssertUnwindSafe(|| f())) {
            Ok(val) => Some(val),
            Err(payload) => {
                self.panicked(payload);
                None
            }
        };

        self.wait_for_jobs(worker);
        self.maybe_propagate_panic();

        // SAFETY: if result panicked, we would have propagated the panic above.
        result.unwrap()
    }

    /// resumes the panic if one happened in this scope.
    fn maybe_propagate_panic(&self) {
        let err_ptr = self.panic.load(Ordering::Relaxed);
        if !err_ptr.is_null() {
            unsafe {
                let err = Box::from_raw(err_ptr);
                std::panic::resume_unwind(*err);
            }
        }
    }

    /// stores the first panic that happened in this scope.
    fn panicked(&self, err: Box<dyn Any + Send + 'static>) {
        self.panic.load(Ordering::Relaxed).is_null().then(|| {
            use core::mem::ManuallyDrop;
            let mut boxed = ManuallyDrop::new(Box::new(err));

            let err_ptr: *mut Box<dyn Any + Send + 'static> = &mut **boxed;
            if self
                .panic
                .compare_exchange(
                    ptr::null_mut(),
                    err_ptr,
                    Ordering::SeqCst,
                    Ordering::Relaxed,
                )
                .is_ok()
            {
                // we successfully set the panic, no need to drop
            } else {
                // drop the error, someone else already set it
                _ = ManuallyDrop::into_inner(boxed);
            }
        });
    }

    pub fn spawn<F>(&'scope self, f: F)
    where
        F: FnOnce(&'scope Self) + Send,
    {
        self.job_counter.increment();

        let this = SendPtr::new_const(self).unwrap();

        let job = Box::new(HeapJob::new(move || unsafe {
            _ = f(this.as_ref());
            this.as_unchecked_ref().job_counter.decrement();
        }))
        .into_boxed_job();

        tracing::trace!("allocated heapjob");

        WorkerThread::current_ref()
            .expect("spawn is run in workerthread.")
            .push_front(job as _);

        tracing::trace!("leaked heapjob");
    }

    pub fn spawn_future<T, F>(&'scope self, future: F) -> async_task::Task<T>
    where
        F: Future<Output = T> + Send + 'scope,
        T: Send + 'scope,
    {
        self.spawn_async_internal(move |_| future)
    }

    #[allow(dead_code)]
    pub fn spawn_async<T, Fut, Fn>(&'scope self, f: Fn) -> async_task::Task<T>
    where
        Fn: FnOnce(&'scope Self) -> Fut + Send + 'scope,
        Fut: Future<Output = T> + Send + 'scope,
        T: Send + 'scope,
    {
        self.spawn_async_internal(f)
    }

    #[inline]
    fn spawn_async_internal<T, Fut, Fn>(&'scope self, f: Fn) -> async_task::Task<T>
    where
        Fn: FnOnce(&'scope Self) -> Fut + Send + 'scope,
        Fut: Future<Output = T> + Send + 'scope,
        T: Send + 'scope,
    {
        self.job_counter.increment();

        let this = SendPtr::new_const(self).unwrap();
        // let this = SendPtr::new_const(&self.job_counter).unwrap();

        let future = async move {
            // SAFETY: this is valid until we decrement the job counter.
            unsafe {
                let _guard = DropGuard::new(move || {
                    this.as_unchecked_ref().job_counter.decrement();
                });
                f(this.as_ref()).await
            }
        };

        let schedule = move |runnable: Runnable| {
            #[align(8)]
            unsafe fn harness(this: *const (), job: *const Job) {
                unsafe {
                    let runnable =
                        Runnable::<()>::from_raw(NonNull::new_unchecked(this.cast_mut()));
                    runnable.run();

                    // SAFETY: job was turned into raw
                    drop(Box::from_raw(job.cast_mut()));
                }
            }

            let job = Box::new(Job::new(harness, runnable.into_raw()));

            // casting into Job<()> here
            WorkerThread::current_ref()
                .expect("spawn_async_internal is run in workerthread.")
                .push_front(Box::into_raw(job) as _);
        };

        let (runnable, task) = unsafe { async_task::spawn_unchecked(future, schedule) };

        runnable.schedule();

        task
    }

    #[inline]
    pub fn join<A, B, RA, RB>(&'scope self, a: A, b: B) -> (RA, RB)
    where
        RA: Send,
        RB: Send,
        A: FnOnce(&'scope Self) -> RA + Send,
        B: FnOnce(&'scope Self) -> RB + Send,
    {
        let worker = WorkerThread::current_ref().expect("join is run in workerthread.");
        let this = SendPtr::new_const(self).unwrap();

        worker.join_heartbeat_every::<_, _, _, _, 64>(
            {
                let this = this;
                move || a(unsafe { this.as_ref() })
            },
            {
                let this = this;
                move || b(unsafe { this.as_ref() })
            },
        )
    }

    unsafe fn from_context(context: Arc<Context>) -> Self {
        Self {
            context,
            job_counter: CountLatch::new(MutexLatch::new()),
            panic: AtomicPtr::new(ptr::null_mut()),
            _scope: PhantomData,
            _env: PhantomData,
        }
    }
}

#[cfg(test)]
mod tests {
    use std::sync::atomic::AtomicU8;

    use tracing_test::traced_test;

    use super::*;
    use crate::ThreadPool;

    #[test]
    fn spawn() {
        let pool = ThreadPool::new_with_threads(1);
        let count = Arc::new(AtomicU8::new(0));

        scope_with_context(&pool.context, |scope| {
            scope.spawn(|_| {
                count.fetch_add(1, std::sync::atomic::Ordering::SeqCst);
            });
        });

        assert_eq!(count.load(std::sync::atomic::Ordering::SeqCst), 1);
    }

    #[test]
    fn join() {
        let pool = ThreadPool::new_with_threads(1);

        let a = pool.scope(|scope| {
            let (a, b) = scope.join(|_| 3 + 4, |_| 5 + 6);
            a + b
        });

        assert_eq!(a, 18);
    }

    #[test]
    fn join_many() {
        let pool = ThreadPool::new_with_threads(1);

        fn sum<'scope, 'env>(scope: &'scope Scope<'scope, 'env>, n: usize) -> usize {
            if n == 0 {
                return 0;
            }

            let (l, r) = scope.join(|s| sum(s, n - 1), |s| sum(s, n - 1));

            l + r + 1
        }

        pool.scope(|scope| {
            let total = sum(scope, 5);
            assert_eq!(total, 31);
            // eprintln!("Total sum: {}", total);
        });
    }

    #[test]
    fn spawn_future() {
        let pool = ThreadPool::new_with_threads(1);
        let mut x = 0;
        pool.scope(|scope| {
            let task = scope.spawn_async(|_| async {
                x += 1;
            });

            task.detach();
        });

        assert_eq!(x, 1);
    }

    #[test]
    fn spawn_many() {
        let pool = ThreadPool::new_with_threads(1);
        let count = Arc::new(AtomicU8::new(0));

        pool.scope(|scope| {
            for _ in 0..10 {
                let count = count.clone();
                scope.spawn(move |_| {
                    count.fetch_add(1, std::sync::atomic::Ordering::SeqCst);
                });
            }
        });

        assert_eq!(count.load(std::sync::atomic::Ordering::SeqCst), 10);
    }
}
